@@ -45,9 +45,9 @@ namespace types
 
   template <>
   struct to_slice<none_type> {
-    using type = contiguous_slice;
+    using type = fast_contiguous_slice;
     static constexpr bool is_new_axis = true;
-    contiguous_slice operator()(none_type);
+    fast_contiguous_slice operator()(none_type);
   };
 
   template <class T>
@@ -116,33 +116,28 @@ namespace types
               std::get<Is>(s).normalize(expr.template shape<Is>())...};
     }
 
-    template <class E, class... S>
-    numpy_gexpr<typename std::decay<E>::type, normalized_slice,
-                normalize_t<S>...>
-    operator()(E &&expr, slice const &s0, S const &... s)
-    {
-      return fwd(std::forward<E>(expr), std::make_tuple(s0, s...),
-                 utils::make_index_sequence<sizeof...(S) + 1>());
-    }
-
-    template <class E, class... S>
-    numpy_gexpr<typename std::decay<E>::type, contiguous_normalized_slice,
-                normalize_t<S>...>
-    operator()(E &&expr, contiguous_slice const &s0, S const &... s)
+    template <class E, class Sp, class... S>
+    typename std::enable_if<
+        is_slice<Sp>::value,
+        numpy_gexpr<typename std::decay<E>::type, normalize_t<Sp>,
+                    normalize_t<S>...>>::type
+    operator()(E &&expr, Sp const &s0, S const &... s)
     {
       return make_gexpr(std::forward<E>(expr), s0, s...);
     }
 
     template <class E, class F, class... S>
-    numpy_gexpr<ndarray<typename std::decay<E>::type::dtype,
-                        array<long, std::decay<E>::type::value>>,
-                contiguous_normalized_slice, normalize_t<S>...>
+    typename std::enable_if<
+        !is_slice<F>::value,
+        numpy_gexpr<ndarray<typename std::decay<E>::type::dtype,
+                            array<long, std::decay<E>::type::value>>,
+                    contiguous_normalized_slice, normalize_t<S>...>>::type
     operator()(E &&expr, F const &s0, S const &... s)
     {
       return numpy_vexpr<ndarray<typename std::decay<E>::type::dtype,
                                  array<long, std::decay<E>::type::value>>,
                          F>{std::forward<E>(expr), s0}(
-          contiguous_slice(none_type{}, none_type{}), s...);
+          fast_contiguous_slice(none_type{}, none_type{}), s...);
     }
   };
 
@@ -549,7 +544,7 @@ namespace types
     // It contains compacted sorted slices value in lower, step && upper is
     // the same as shape.
     // indices for long index are store in the indices array.
-    // position for slice && long value in the extended slice can be found
+    // position for slice and long value in the extended slice can be found
     // through the S... template
     // && compacted values as we know that first S is a slice.
 
@@ -592,7 +587,7 @@ namespace types
                                   const_nditerator<numpy_gexpr>,
                                   dtype const *>::type;
 
-    typename std::remove_reference<Arg>::type arg;
+    Arg arg;
 
     std::tuple<S...> slices;
 
@@ -706,6 +701,9 @@ namespace types
 
     numpy_gexpr &operator=(numpy_gexpr const &expr);
 
+    template <class Argp>
+    numpy_gexpr &operator=(numpy_gexpr<Argp, S...> const &expr);
+
     template <class Op, class E>
     typename std::enable_if<may_overlap_gexpr<E>::value, numpy_gexpr &>::type
     update_(E const &expr);
@@ -767,11 +765,12 @@ namespace types
       return numpy_iexpr_helper<numpy_gexpr, value>::get(*this, i);
     }
 
-    template <class... Indices>
-    void store(dtype elt, Indices... indices)
+    template <class E, class... Indices>
+    void store(E elt, Indices... indices)
     {
+      static_assert(is_dtype<E>::value, "valid store");
       *(buffer + noffset<value>{}(*this, array<long, value>{{indices...}})) =
-          elt;
+          static_cast<E>(elt);
     }
     template <class... Indices>
     dtype load(Indices... indices) const
@@ -779,12 +778,13 @@ namespace types
       return *(buffer +
                noffset<value>{}(*this, array<long, value>{{indices...}}));
     }
-    template <class Op, class... Indices>
-    void update(dtype elt, Indices... indices) const
+    template <class Op, class E, class... Indices>
+    void update(E elt, Indices... indices) const
     {
+      static_assert(is_dtype<E>::value, "valid store");
       Op{}(
           *(buffer + noffset<value>{}(*this, array<long, value>{{indices...}})),
-          elt);
+          static_cast<E>(elt));
     }
 
 #ifdef USE_XSIMD
@@ -799,10 +799,9 @@ namespace types
     template <class... Sp>
     auto operator()(Sp const &... s) const -> decltype(make_gexpr(*this, s...));
 
-    auto operator[](contiguous_slice const &s0) const
-        -> decltype(make_gexpr(*this, s0));
-
-    auto operator[](slice const &s0) const -> decltype(make_gexpr(*this, s0));
+    template <class Sp>
+    auto operator[](Sp const &s) const -> typename std::enable_if<
+        is_slice<Sp>::value, decltype(make_gexpr(*this, (s.lower, s)))>::type;
 
     template <size_t M>
     auto fast(array<long, M> const &indices) const
@@ -870,19 +869,14 @@ namespace types
   };
 }
 
-template <class Arg, class... S>
-struct returnable<types::numpy_gexpr<Arg, S...>> {
-  using type = types::numpy_gexpr<typename returnable<Arg>::type, S...>;
-};
-
 template <class T, class pS, class... S>
 struct assignable<types::numpy_gexpr<types::ndarray<T, pS> const &, S...>> {
-  using type = types::numpy_gexpr<types::ndarray<T, pS> const &, S...>;
+  using type = types::numpy_gexpr<types::ndarray<T, pS>, S...>;
 };
 
 template <class T, class pS, class... S>
 struct assignable<types::numpy_gexpr<types::ndarray<T, pS> &, S...>> {
-  using type = types::numpy_gexpr<types::ndarray<T, pS> &, S...>;
+  using type = types::numpy_gexpr<types::ndarray<T, pS>, S...>;
 };
 
 template <class Arg, class... S>
@@ -891,16 +885,10 @@ struct assignable<types::numpy_gexpr<Arg, S...>> {
 };
 
 template <class Arg, class... S>
-struct lazy<types::numpy_gexpr<Arg, S...>> {
-  using type =
-      types::numpy_gexpr<typename lazy<Arg>::type, typename lazy<S>::type...>;
+struct lazy<types::numpy_gexpr<Arg, S...>>
+    : assignable<types::numpy_gexpr<Arg, S...>> {
 };
 
-template <class Arg, class... S>
-struct lazy<types::numpy_gexpr<Arg const &, S...>> {
-  using type = types::numpy_gexpr<typename lazy<Arg>::type const &,
-                                  typename lazy<S>::type...>;
-};
 PYTHONIC_NS_END
 /* type inference stuff  {*/
 #include "pythonic/include/types/combined.hpp"
